# coding: utf-8

from __future__ import absolute_import
from __future__ import unicode_literals

from uuid import uuid4

from django.conf import settings
from django.contrib.auth import login
from django.contrib.auth import logout
from django.contrib.auth import get_user_model

from m3_auth.proxy import SessionProxy
from m3_auth.utils import import_object
from m3_auth.utils import is_authenticated
from m3_auth.utils import get_request_params
from m3_auth.exceptions import UserNotFound
from m3_auth.exceptions import AuthError
from m3_auth.exceptions import BadRequest
from m3_auth.exceptions import MoreUsersFound
from m3_auth.backends.base import BaseBackend
from m3_auth.backends.base.remote.signals import user_not_found
from m3_auth.backends.base.remote.signals import user_authenticated


class RemoteBackend(BaseBackend):

    """
    Базовый бэкенд для обработки запросов на вход и выход через сторонние 
    сервера авторизаций по протоколам OAuth2.0 и SAML2.
    
    Принцип работы данного бэкенда следующий:
    
    Вход:
    1. Пользователь переадресовывается на сервер авторизации по полученному URL
    методом get_login_url.
    2. Вводит свои данные, дает разрешение на передачу своих данных и сервер 
    авторизации возвращает его обратно в систему.
    3. Осуществляется обработка полученного запроса на вход методом 
    process_login, process_login передает управление объекту протоколу,
    который обрабатывает сам запрос и возвращает объект класса с информацией о 
    пользователе.
    4. Далее, применяются фильтры из user_filters для поиска подходящих 
    пользователей и осуществляется вход.
    
    Выход:
    1. Пользователь переадресовывается на URL адрес полученный методом 
    get_logout_url, это может быть URL адрес сервера авторизации, либо адрес 
    в системе который обрабатывает выход.
    2. Обрабатывается запрос методом process_logout и осуществляется выход.
    
    :param protocol: Объект класса конкретного протокола.
    :type protocol: auth.protocols.base.BaseProtocol

    :param user_filters: Список фильтров для поиска подходящих пользователей. 
        Применяется после успешной обработки запроса и получения данных 
        пользователя. Конкретные фильтры реализуются в проектах под конкретные 
        требования сопоставления пользователей. Фильтры это обычные callable 
        объекты python, которые должны принимать:
        1. request - запрос.
        2. queryset - queryset пользователей.
        3. remote_user - объект класса пользователя с сервера авторизации.
    
    :type user_filters: list of callable
    """

    def __init__(self, name, protocol, user_filters):
        super(RemoteBackend, self).__init__(name)
        self.protocol = protocol
        self.filters = user_filters

    def get_login_url(self, request):
        """Метод возвращает URL адрес, на который необходимо переадресовать
        пользователя для аутентификации через сторонний сервер авторизации.
        Возвращаемый URL адрес зависит от протокола.

        :param request: Запрос
        :type request: django.http.request.HttpRequest

        :return: URL Адрес.
        :rtype: str
        """
        return self.protocol.get_login_url(request)

    def process_login(self, request):
        """Метод обрабатывает запрос на вход, через сторонний сервер 
        авторизации. Находит подходящих пользователей применяя фильтрацию
        из списка фильтров, и найдя одного пользователя аутентифицирует его.

        Если пользователей не найдено, то будет исключение UserNotFound.
        Если пользователей найдено несколько, то будет исключение 
        MoreUsersFound. При этом, список id пользователей будет сохранен в 
        сессии для возможности предоставления выбора.

        :param request: Запрос.
        :type request: django.http.request.HttpRequest

        :raises: UserNotFound, MoreUsersFound, AuthError

        :return: Объект класса пользователь.
        :rtype: django.contrib.auth.models.User
        """
        session_proxy = SessionProxy(request)
        remote_user = self.protocol.process_login(request)
        session_proxy.remote_user_id = remote_user.user_id

        users = get_user_model().objects
        users = self.apply_filters(request, users, remote_user) if (
            self.filters) else users.none()

        users_count = users.count()

        if users_count == 0:
            user_not_found.send(
                type(self), request=request, remote_user=remote_user)
            raise UserNotFound('Пользователь не найден в системе.')
        elif users_count > 1:
            session_proxy.token = str(uuid4())
            session_proxy.remote_user_id = remote_user.user_id
            session_proxy.user_ids_to_select = tuple(u.id for u in users)
            raise MoreUsersFound('Нельзя однозначно определить пользователя.')

        selected_user = users.first()
        self._login(request, selected_user, remote_user.user_id)
        user_authenticated.send(
            type(self), request=request, remote_user=remote_user)
        return selected_user

    def get_logout_url(self, request):
        """Метод возвращает URL адрес, на который необходимо переадресовать
        пользователя для того чтобы выйти из системы. URL адрес зависит от 
        протокола.

        :param request: Запрос
        :type request: django.http.request.HttpRequest

        :return: URL Адрес.
        :rtype: str
        """
        return self.protocol.get_logout_url(request)

    def process_logout(self, request):
        """Метод обрабатывает запрос и осуществляет выход из системы.

        :param request: Запрос
        :type request: django.http.request.HttpRequest
        :rtype: bool
        """
        if self.protocol.process_logout(request):
            self._logout(request)
            return True

    def login_with_selected_user(self, request):
        """Метод обрабатывает запрос на вход, под выбранным пользователем.
        
        Ожидаются следующие параметры запроса:
        1. user_id - Идентификатор выбранного пользователя.
        2. token - временный токен подтверждающий о предоставленном выборе.
        (Перед предоставлением выбора сохраняется в сессию, получить его можно 
        из SessionProxy)

        :param request: Запрос
        :type request: django.http.request.HttpRequest
        
        :raises: auth.exceptions.AuthError, auth.exceptions.BadRequest
        
        :return: Объект класса пользователь.
        :rtype: django.contrib.auth.models.User
        """
        request_params = get_request_params(request)
        session_proxy = SessionProxy(request)
        session_token = session_proxy.token
        session_users = session_proxy.user_ids_to_select

        if session_token is None:
            raise BadRequest()
        if is_authenticated(request):
            raise AuthError('Вы уже аутентифицированы.')

        # Токен выдается только на одну попытку
        session_proxy.token = None
        session_proxy.user_ids_to_select = None

        token = request_params.get('token')
        user_id = request_params.get('user_id')

        if not all((token, user_id)) or not user_id.isdigit():
            raise BadRequest()

        token_is_correct = token == session_token
        user = get_user_model().objects.filter(pk=user_id).first()
        if not all((token_is_correct, user)) or user.pk not in session_users:
            raise BadRequest()

        if not user.is_active:
            raise AuthError('Пользователь "{}" не активен.'.format(
                user.username))

        self._login(request, user, session_proxy.remote_user_id)
        return user

    def apply_filters(self, request, queryset, remote_user):
        """Метод применяет фильтрацию для поиска подходящих пользователей.
        Если указанных фильтров нет - возвращается пустой QuerySet во избежание
        получения всех пользователей.
        
        :param request: Запрос
        :type request: django.http.request.HttpRequest

        :param queryset: QuerySet к которому необходимо применить фильтрацию.
        :type queryset: django.db.models.query.QuerySet

        :param remote_user: Объект класса представляющего пользователя с 
            сервера авторизации.
        :type remote_user: auth.backends.base.remote.user.BaseRemoteUser

        :return: Отфильтрованный QuerySet
        :rtype: django.db.models.query.QuerySet
        """
        if not self.filters:
            return queryset.none()

        for _filter in self.filters:
            queryset = _filter(request, queryset, remote_user)
        return queryset

    def _login(self, request, user, remote_user_id):
        """Аутентификация пользователя в системе без пароля.
        Функция симулирует работу django.contrib.auth.authenticate().

        :param request: Запрос
        :type request: django.http.request.HttpRequest 

        :param user: Пользователь системы.
        :type user: django.contrib.auth.models.User

        :param remote_user_id: Идентификатор пользователя системы авторизации.
        :return: None
        """
        user.backend = settings.AUTHENTICATION_BACKENDS[0]
        login(request, user)

        session_proxy = SessionProxy(request)
        session_proxy.remote_user_id = remote_user_id
        session_proxy.backend = self.name

    def _logout(self, request):
        """Метод осуществляет выход из системы.

        :param request: Запрос
        :type request: django.http.request.HttpRequest
        :return: None
        """
        logout(request)

    @classmethod
    def from_config(cls, config):
        """Метод создает объект бэкенда на основе объекта конфигураций.
        
        Обязательные параметры config:
        1. name - Наименование бэкенда.
        2. user_filters - список ссылок/путей до функций/callable объектов 
        реализующих фильтрацию. (н-р. ["project.filters.fullname_filter"]).
        3. protocol - Настройки используемого протокола.
        4. protocol.protocol_cls - Путь/ссылка на класс конкретного протокола.
        (н-р, "auth.backends.oauth.protocol.OAuthProtocol")
        
        :param config: Объект настроек.
        :type config: auth.config.Config

        :return: Объект бэкенда.
        :rtype: auth.backends.base.remote.RemoteBackend
        """

        # Получаем класс протокола
        protocol_cls = import_object(config.protocol.protocol_cls)
        # Создаем объект протокола из настроек.
        protocol = protocol_cls.from_config(config.protocol)

        user_filters = []
        for user_filter in config.user_filters:
            user_filters.append(import_object(user_filter))

        return cls(config.name, protocol, user_filters)
