# coding: utf-8

from __future__ import absolute_import
from __future__ import unicode_literals

from django.contrib.auth import get_user_model


class DefaultChecker(object):

    """
    Класс для проверки переданных данных на аутентификацию пользователя.
    Используется по умолчанию.
    
    :param request: Запрос.
    :type request: django.http.HttpRequest
    
    :param username: Логин пользователя, либо значение USERNAME_FIELD из 
        настроек модели User. Например, если аутентификация осуществляется по 
        email'у и паролю.
    :type username: str or unicode
    
    :param password: Пароль пользователя.
    :type password: str or unicode
    """

    USER_NOT_EXISTS = 'Пользователь не найден.'
    INACTIVE_USER = 'Пользователь заблокирован.'
    INCORRECT_PASSWORD = 'Неверный {} или пароль.'

    def __init__(self, request, username, password):
        self.request = request
        self.username = username
        self.password = password

        self.user_model = get_user_model()
        self.username_field = getattr(
            self.user_model, 'USERNAME_FIELD', 'username')

        self.user = self.user_model.objects.filter(**{
            self.username_field: self.username}).first()

    def check_if_exists(self):
        """Проверяет существование пользователя."""
        if not self.user:
            return self.USER_NOT_EXISTS

    def check_if_active(self):
        """Проверяет, активен ли пользователь или нет."""
        if not self.user.is_active:
            return self.INACTIVE_USER

    def check_password(self):
        """Проверяет правильность пароля пользователя."""
        if not self.user.check_password(self.password):
            field = get_user_model()._meta.get_field(self.username_field)
            return self.INCORRECT_PASSWORD.format(field.verbose_name)

    def check(self):
        """Проверяет данные для аутентификации.
        :return: Сообщение об ошибке, либо None если все хорошо.
        """
        for method in [
            self.check_if_exists,
            self.check_if_active,
            self.check_password
        ]:
            error = method()
            if error:
                return error
