# coding: utf-8

from __future__ import absolute_import
from __future__ import unicode_literals

import six

from onelogin.saml2.auth import OneLogin_Saml2_Auth as SAMLAuth
from onelogin.saml2.utils import OneLogin_Saml2_Utils as SAMLUtils
from onelogin.saml2.errors import OneLogin_Saml2_Error as SAMLError
from onelogin.saml2.settings import OneLogin_Saml2_Settings as SAMLSettings
from onelogin.saml2.logout_request import (
    OneLogin_Saml2_Logout_Request as LogoutRequest)

from m3_auth.exceptions import AuthError, BadRequest
from m3_auth.utils import get_request_params
from m3_auth.utils import import_object


class SAMLProtocol(object):

    """
    Класс реализует формирование URL адресов, обработку запросов на 
    аутентификацию, Single logout по протоколу SAML 2.
    
    :param user_cls: Класс который будет представлять информацию о пользователе 
        с сервера авторизации.
    :type user_cls: type
    
    :param config: Настройки протокола.
    :type config: auth.config.Config
    """

    def __init__(self, user_cls, config):
        super(SAMLProtocol, self).__init__()
        self.user_cls = user_cls
        self.config = config

    def get_login_url(self, request):
        """Метод формирует и возвращает URL адрес для переадресации на сервер 
        авторизации.

        :param request: django.http.request.HttpRequest
        :return: URL адрес.
        :rtype: str or unicode
        """
        return self._get_auth(request).login()

    def process_login(self, request):
        """Обрабатывает запрос на аутентификацию по протоколу SAML, извлекает 
        информацию и возвращает объект пользователя.
        
        В случае возникновения ошибок, будет выброшено исключение AuthError.
        
        :param request: Запрос
        :type request: django.http.request.HttpRequest
        :raises: auth.exceptions.AuthError

        :return: Объект представляющий информацию о пользователе.
        :rtype: auth.backends.saml2.user.BaseSAMLUser
        """
        auth = self._get_auth(request)

        try:
            auth.process_response()
        except SAMLError as e:
            raise AuthError(six.text_type(e))

        if auth.get_errors():
            raise AuthError(auth.get_last_error_reason())

        user_attributes = {'user_id': auth.get_nameid()}
        user_attributes.update(auth.get_attributes())
        return self.user_cls(user_attributes)

    def get_logout_url(self, request):
        """Формирует и возвращает URL адрес для перенаправления пользователя 
        на сервер авторизации для выхода.

        :param request: Запрос
        :type request: django.http.request.HttpRequest

        :return: URL адрес.
        :rtype: str or unicode
        """
        return self._get_auth(request).logout(
            self.config.logout_redirect_url)

    def process_logout(self, request):
        """Обрабатывает запрос на выход по протоколу SAML и возвращает True 
        если все без ошибок, в противном случае исключение AuthError.

        :param request: Запрос
        :type request: django.http.request.HttpRequest

        :raises: auth.exceptions.AuthError
        :return: True или False
        :rtype: bool
        """
        auth = self._get_auth(request)
        auth.process_slo()

        if auth.get_errors():
            raise AuthError(auth.get_last_error_reason())

        return True

    def process_single_logout(self, request):
        """Обрабатывает запрос на выход из системы.

        Реализует поддержку Single Logout: если пользователь вошел
        через сервер авторизации в другие системы, а затем вышел из одной из 
        них, сервер авторизации отправит всем остальным системам запрос на 
        прекращение сессии.

        :param request: Запрос
        :type request: django.http.request.HttpRequest

        :return: user_id, response redirect_url
        :rtype: tuple
        """
        auth = self._get_auth(request)
        request_params = get_request_params(request)

        if 'SAMLRequest' not in request_params:
            raise BadRequest()

        # Обработка запроса
        saml_request = request_params['SAMLRequest']
        saml_settings = SAMLSettings(
            custom_base_path=self.config.custom_base_path)
        logout_request = LogoutRequest(saml_settings, saml_request)

        request_data = self._get_request_data(request)
        if not logout_request.is_valid(request_data):
            raise AuthError(logout_request.get_error())

        request_xml = SAMLUtils.decode_base64_and_inflate(saml_request)

        user_id = LogoutRequest.get_nameid(request_xml)
        request_id = LogoutRequest.get_id(request_xml)
        return user_id, auth.process_slo(request_id=request_id)

    def _get_auth(self, request):
        """Создает и возращает объект SP SAML.

        :param request: django.http.request.HttpRequest
        :return: Объект OneLogin_Saml2_Auth
        :rtype: onelogin.saml2.auth.OneLogin_Saml2_Auth
        """
        return SAMLAuth(
            custom_base_path=self.config.custom_base_path,
            request_data=self._get_request_data(request)
        )

    @staticmethod
    def _get_request_data(request):
        """Подготавливает и возвращает словарь параметров для всех запросов.

        :param request: Запрос
        :type request: django.http.request.HttpRequest

        :return: Словарь параметров для запросов SAML
        :rtype: dict
        """
        return dict(
            https='on' if request.is_secure() else 'off',
            http_host=request.META['HTTP_HOST'],
            script_name=request.META['PATH_INFO'],
            server_port=request.META['SERVER_PORT'],
            get_data=request.GET.copy(),
            post_data=request.POST.copy(),
        )

    @classmethod
    def from_config(cls, config):
        user_class = import_object(config.user_cls)
        return cls(user_class, config.config)
